import json

from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponseBadRequest, JsonResponse
from django.shortcuts import redirect, render
from django.utils.html import format_html
from django.views.decorators.csrf import csrf_exempt

from eveuniverse.models import EveType

from ..forms import NotificationForm
from ..models import Notification, Program, ProgramLocation
from ..utils import MessagesPlus


@csrf_exempt
@login_required
@permission_required("buybacks2.basic_access")
def program_notify(request, program_pk):
    program = Program.objects.filter(pk=program_pk).first()

    if request.method != "POST" or program is None:
        return HttpResponseBadRequest("")

    data = json.loads(request.body)

    program_location = ProgramLocation.objects.filter(
        program=program, id=data["program_location"]
    ).first()

    if program_location is None:
        return HttpResponseBadRequest("")

    notification = Notification.objects.create(
        program_location=program_location,
        user=request.user,
        total=data["total"],
        items=json.dumps(data["items"]),
    )

    if notification is None:
        return HttpResponseBadRequest("")
    else:
        MessagesPlus.success(
            request,
            format_html(
                "Created a notification for buyback program <strong>{}</strong>",
                program.name,
            ),
        )

    return JsonResponse({})


@login_required
@permission_required("buybacks2.basic_access")
def my_notifications(request):
    notifications = Notification.objects.filter(
        user=request.user,
    )

    items = process_notifications(notifications)

    context = {
        "bb_notifications": notifications,
        "items": items,
        "mine": True,
    }

    return render(request, "buybacks2/notifications.html", context)


@login_required
@permission_required("buybacks2.basic_access")
def notification_remove(request, notification_pk):
    mine = request.GET.get("mine", "True") == "True"

    notification = Notification.objects.filter(pk=notification_pk)

    if mine:
        notification.filter(user=request.user).delete()
    elif request.user.has_perm("buybacks2.manage_programs"):
        notification = notification.first()
        notification.delete()

        return redirect(
            "buybacks2:program_notifications",
            program_pk=notification.program_location.program.id,
        )

    return redirect("buybacks2:my_notifications")


@login_required
@permission_required("buybacks2.basic_access")
def notification_edit(request, notification_pk):
    mine = request.GET.get("mine", "True") == "True"

    notification = Notification.objects.filter(pk=notification_pk)

    if not mine and not request.user.has_perm("buybacks2.manage_programs"):
        return redirect("buybacks2:index")

    if mine:
        notification = notification.filter(user=request.user)

    notification = notification.first()

    if notification is None:
        return redirect("buybacks2:my_notifications")

    if request.method != "POST":
        form = NotificationForm(notification=notification)
    else:
        form = NotificationForm(request.POST, notification=notification)

        if form.is_valid():
            notification.program_location = form.cleaned_data["office"]

            try:
                notification.save()

                MessagesPlus.success(
                    request,
                    format_html(
                        "Edited notification location to <strong>{}</strong>",
                        notification.program_location,
                    ),
                )

                if mine:
                    return redirect("buybacks2:my_notifications")
                else:
                    return redirect(
                        "buybacks2:program_notifications",
                        program_pk=notification.program_location.program.id,
                    )

            except Exception:
                MessagesPlus.error(
                    request,
                    "Failed to edit location of the notification",
                )

    context = {
        "corporation": notification.program_location.program.corporation.corporation,
        "form": form,
        "notification": notification,
        "mine": mine,
    }

    return render(request, "buybacks2/notification_edit.html", context)


@login_required
@permission_required("buybacks2.basic_access")
def program_notifications(request, program_pk):
    notifications = Notification.objects.filter(
        program_location__program__id=program_pk,
    )

    items = process_notifications(notifications)

    context = {
        "bb_notifications": notifications,
        "items": items,
        "mine": False,
    }

    return render(request, "buybacks2/notifications.html", context)


def process_notifications(notifications: dict) -> dict:
    items = {}
    typeids = set()
    for notification in notifications:
        data = json.loads(notification.items)

        for type_id in data:
            typeids.add(type_id)
    types = EveType.objects.filter(pk__in=typeids)
    for item in types:
        items[str(item.id)] = item.name
    return items



from django import forms

from eveuniverse.models import EveType

from .models import Office, ProgramLocation
from .validators import validate_brokerage


class ProgramForm(forms.Form):
    name = forms.CharField(
        label="Program Name",
        help_text="Give your buyback program a name to easily identify it",
    )

    def __init__(self, *args, **kwargs):
        program = kwargs.pop("program", None)

        super().__init__(*args, **kwargs)

        if program is not None:
            self.fields["name"].initial = program.name


class ProgramItemForm(forms.Form):
    item_type = forms.ModelChoiceField(
        queryset=EveType.objects.none(),
        label="Item type",
        help_text="Add the name of item which is accepted in this buyback program. Once you start typing, we offer suggestions",
        empty_label=None,
    )
    brokerage = forms.IntegerField(
        label="Brokerage",
        help_text="Set the percentage corporation gets on top of 'Jita Max Buy'",
        validators=[validate_brokerage],
    )
    # use_refined_value = forms.BooleanField(
    #     label="Calculate value using the refined end product (only works for ore)",
    #     required=False,
    # )

    def __init__(self, *args, **kwargs):
        value = kwargs.pop("value", None)

        super().__init__(*args, **kwargs)

        if value is not None:
            self.fields["item_type"].queryset = EveType.objects.filter(
                pk=value,
                published=True,
            ).exclude(eve_group__eve_category__id=9)


class ProgramLocationForm(forms.Form):
    office = forms.ModelChoiceField(
        queryset=Office.objects.none(),
        label="Location",
        help_text="Add the location where buybacks2 in this program are accepted",
        empty_label=None,
    )

    def __init__(self, *args, **kwargs):
        program = kwargs.pop("program", None)

        super().__init__(*args, **kwargs)

        self.fields["office"].queryset = Office.objects.filter(
            corporation=program.corporation,
        )


class CalculatorForm(forms.Form):
    office = forms.ModelChoiceField(
        queryset=ProgramLocation.objects.none(),
        label="Location",
        help_text="Structure or station from where you want to sell items",
        empty_label=None,
    )
    items = forms.CharField(
        widget=forms.Textarea,
        label="Items",
        help_text="Copy and paste the item data from your inventory. Item types not in this buyback program will be ignored",
    )

    def __init__(self, *args, **kwargs):
        program = kwargs.pop("program", None)

        super().__init__(*args, **kwargs)

        self.fields["office"].queryset = ProgramLocation.objects.filter(
            program=program,
        )


class NotificationForm(forms.Form):
    user = forms.CharField(
        label="User",
        help_text="User who created this notification",
    )
    office = forms.ModelChoiceField(
        queryset=ProgramLocation.objects.none(),
        label="Location",
        help_text="Structure or station from where you contracted your items",
        empty_label=None,
    )

    def __init__(self, *args, **kwargs):
        notification = kwargs.pop("notification", None)

        super().__init__(*args, **kwargs)

        self.fields["user"].initial = notification.user
        self.fields["user"].widget.attrs["readonly"] = True

        self.fields["office"].queryset = ProgramLocation.objects.filter(
            program=notification.program_location.program,
        )
        self.fields["office"].initial = notification.program_location



from django.contrib.auth.decorators import login_required, permission_required
from django.db import transaction
from django.db.models import F
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.html import format_html
from django.utils.translation import gettext_lazy

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from esi.decorators import token_required
from eveuniverse.models import EveType

from ..models import Corporation, Program
from ..tasks import update_offices_for_corp
from ..utils import MessagesPlus


@login_required
def index(request):
    context = {"programs": Program.objects.filter()}

    return render(request, "buybacks2/index.html", context)


@login_required
@permission_required("buybacks2.manage_programs")
def item_autocomplete(request):
    items = EveType.objects.filter(published=True).exclude(
        eve_group__eve_category__id=9
    )

    q = request.GET.get("q", None)

    if q is not None:
        items = items.filter(name__icontains=q)

    items = items.annotate(
        value=F("id"),
        text=F("name"),
    ).values("value", "text")

    return JsonResponse(list(items), safe=False)


@login_required
@permission_required("buybacks2.setup_retriever")
@token_required(
    scopes=[
        "esi-universe.read_structures.v1",
        "esi-assets.read_corporation_assets.v1",
        "esi-contracts.read_corporation_contracts.v1",
    ]
)
def setup(request, token):
    success = True
    owned_char = None
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    try:
        owned_char = CharacterOwnership.objects.get(
            user=request.user, character=token_char
        )
    except CharacterOwnership.DoesNotExist:
        MessagesPlus.error(
            request,
            format_html(
                gettext_lazy(
                    "You can only use your main or alt characters to add corporations. "
                    "However, character %s is neither. "
                )
                % format_html("<strong>{}</strong>", token_char.character_name)
            ),
        )
        success = False

    if success:
        try:
            corporation = EveCorporationInfo.objects.get(
                corporation_id=token_char.corporation_id
            )
        except EveCorporationInfo.DoesNotExist:
            corporation = EveCorporationInfo.objects.create_corporation(
                token_char.corporation_id
            )

        with transaction.atomic():
            corp, _ = Corporation.objects.update_or_create(
                corporation=corporation, character=owned_char
            )

            corp.save()

        update_offices_for_corp.delay(corp_pk=corp.pk)

        MessagesPlus.info(
            request,
            format_html(
                gettext_lazy(
                    "%(corporation)s has been added with %(character)s "
                    "as sync character. We have started fetching offices "
                    "for this corporation. You will receive a report once "
                    "the process is finished."
                )
                % {
                    "corporation": format_html("<strong>{}</strong>", corp),
                    "character": format_html(
                        "<strong>{}</strong>", corp.character.character.character_name
                    ),
                }
            ),
        )

    return redirect("buybacks2:index")



EVE_CATEGORY_ID_MATERIAL = 4
EVE_CATEGORY_ID_SHIP = 6
EVE_CATEGORY_ID_MODULE = 7
EVE_CATEGORY_ID_CHARGE = 8
EVE_CATEGORY_ID_COMMODITY = 17
EVE_CATEGORY_ID_DRONE = 18
EVE_CATEGORY_ID_ASTEROID = 25
EVE_CATEGORY_ID_PLANETARY_COMMODITY = 43
EVE_CATEGORY_ID_FIGHTER = 87

EVE_GROUP_ID_HARVESTABLE_CLOUD = 711



from bravado.exception import HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
from celery import shared_task

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from .models import Corporation

DEFAULT_TASK_PRIORITY = 6
TASKS_TIME_LIMIT = 7200

logger = get_extension_logger(__name__)

# Create your tasks here
TASK_DEFAULT_KWARGS = {
    "time_limit": TASKS_TIME_LIMIT,
}

TASK_ESI_KWARGS = {
    **{
        "bind": True,
        "autoretry_for": (
            OSError,
            HTTPBadGateway,
            HTTPGatewayTimeout,
            HTTPServiceUnavailable,
        ),
        "retry_kwargs": {"max_retries": 3},
        "retry_backoff": 30,
    },
}


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "base": QueueOnce,
            "once": {"keys": ["corp_pk"], "graceful": True},
            "max_retries": None,
        },
    }
)
def update_offices_for_corp(self, corp_pk):
    """fetches all office locations for corp from ESI"""
    return _get_corp(corp_pk).update_offices_esi()


@shared_task(**TASK_DEFAULT_KWARGS)
def update_all_offices():
    for corp in Corporation.objects.all():
        update_offices_for_corp.apply_async(
            kwargs={"corp_pk": corp.pk},
            priority=DEFAULT_TASK_PRIORITY,
        )


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "base": QueueOnce,
            "once": {"keys": ["corp_pk"], "graceful": True},
            "max_retries": None,
        },
    }
)
def sync_contracts_for_corp(self, corp_pk):
    """fetches all contracts for corp from ESI"""
    return _get_corp(corp_pk).sync_contracts_esi()


@shared_task(**TASK_DEFAULT_KWARGS)
def sync_all_contracts():
    for corp in Corporation.objects.all():
        sync_contracts_for_corp.apply_async(
            kwargs={"corp_pk": corp.pk},
            priority=DEFAULT_TASK_PRIORITY,
        )


def _get_corp(corp_pk: int) -> Corporation:
    """returns the corp or raises exception"""
    try:
        corp = Corporation.objects.get(pk=corp_pk)
    except Corporation.DoesNotExist:
        raise Corporation.DoesNotExist(
            f"Requested corp with pk {corp_pk} does not exist"
        )
    return corp



from django.conf.urls import include, url

from .views import common, notifications, programs, stats

app_name = "buybacks2"

module_urls = [
    url("", common.index, name="index"),
    url("^setup/", common.setup, name="setup"),
    url(
        "^item_autocomplete/",
        common.item_autocomplete,
        name="item_autocomplete",
    ),
    url("my_notifications/", notifications.my_notifications, name="my_notifications"),
    url("my_stats/", stats.my_stats, name="my_stats"),
    url(
        r"^notification/(?P<notification_pk>[0-9]+)/remove$",
        notifications.notification_remove,
        name="notification_remove",
    ),
    url(
        r"^notification/(?P<notification_pk>[0-9]+)/edit$",
        notifications.notification_edit,
        name="notification_edit",
    ),
    url("program_add/", programs.program_add, name="program_add"),
    url("program_add_2/", programs.program_add_2, name="program_add_2"),
    url(
        r"^program/(?P<program_pk>[0-9]+)/edit$",
        programs.program_edit,
        name="program_edit",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)$",
        programs.program,
        name="program",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/remove$",
        programs.program_remove,
        name="program_remove",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/calculate$",
        programs.program_calculate,
        name="program_calculate",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/stats$",
        stats.program_stats,
        name="program_stats",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/notify$",
        notifications.program_notify,
        name="program_notify",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/notifications$",
        notifications.program_notifications,
        name="program_notifications",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/add_item$",
        programs.program_add_item,
        name="program_add_item",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/add_location$",
        programs.program_add_location,
        name="program_add_location",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/remove_item/(?P<item_type_pk>[0-9]+)$",
        programs.program_remove_item,
        name="program_remove_item",
    ),
    url(
        r"^program/(?P<program_pk>[0-9]+)/remove_location/(?P<office_pk>[0-9]+)$",
        programs.program_remove_location,
        name="program_remove_location",
    ),
]

urlpatterns = [
    url(f"^{app_name}/", include((module_urls, app_name), namespace=app_name)),
]



from django.core.management import call_command
from django.core.management.base import BaseCommand

from ... import __title__
from ...constants import (
    EVE_CATEGORY_ID_ASTEROID,
    EVE_CATEGORY_ID_CHARGE,
    EVE_CATEGORY_ID_COMMODITY,
    EVE_CATEGORY_ID_DRONE,
    EVE_CATEGORY_ID_FIGHTER,
    EVE_CATEGORY_ID_MATERIAL,
    EVE_CATEGORY_ID_MODULE,
    EVE_CATEGORY_ID_PLANETARY_COMMODITY,
    EVE_CATEGORY_ID_SHIP,
    EVE_GROUP_ID_HARVESTABLE_CLOUD,
)


class Command(BaseCommand):
    help = "Preloads data required for aa-buybacks2 from ESI"

    def handle(self, *args, **options):
        call_command(
            "eveuniverse_load_types",
            __title__,
            "--category_id",
            str(EVE_CATEGORY_ID_MATERIAL),
            "--category_id",
            str(EVE_CATEGORY_ID_SHIP),
            "--category_id",
            str(EVE_CATEGORY_ID_MODULE),
            "--category_id",
            str(EVE_CATEGORY_ID_CHARGE),
            "--category_id",
            str(EVE_CATEGORY_ID_COMMODITY),
            "--category_id",
            str(EVE_CATEGORY_ID_DRONE),
            "--category_id",
            str(EVE_CATEGORY_ID_ASTEROID),
            "--category_id",
            str(EVE_CATEGORY_ID_PLANETARY_COMMODITY),
            "--category_id",
            str(EVE_CATEGORY_ID_FIGHTER),
            "--group_id",
            str(EVE_GROUP_ID_HARVESTABLE_CLOUD),
        )



from bravado.exception import HTTPForbidden, HTTPUnauthorized

from django.db import models

from allianceauth.services.hooks import get_extension_logger
from esi.models import Token
from eveuniverse.models import EveSolarSystem

from .helpers import esi_fetch

logger = get_extension_logger(__name__)


class LocationManager(models.Manager):
    STATION_ID_START = 60000000
    STATION_ID_END = 69999999

    def get_or_create_from_esi(
        self, token: Token, location_id: int, add_unknown: bool = True
    ) -> tuple:
        """gets or creates location object with data fetched from ESI"""
        from .models import Location

        try:
            location = self.get(id=location_id)
            created = False
        except Location.DoesNotExist:
            location, created = self.update_or_create_from_esi(
                token=token, location_id=location_id, add_unknown=add_unknown
            )

        return location, created

    def update_or_create_from_esi(
        self, token: Token, location_id: int, add_unknown: bool = True
    ) -> tuple:
        """updates or creates location object with data fetched from ESI"""
        from .models import Location

        if location_id >= self.STATION_ID_START and location_id <= self.STATION_ID_END:
            logger.info("Fetching station from ESI")
            try:
                station = esi_fetch(
                    "Universe.get_universe_stations_station_id",
                    args={"station_id": location_id},
                )

                eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
                    id=station["system_id"]
                )

                location, created = self.update_or_create(
                    id=location_id,
                    defaults={
                        "name": station["name"],
                        "eve_solar_system": eve_solar_system,
                        "category_id": Location.CATEGORY_STATION_ID,
                    },
                )
            except Exception as ex:
                logger.exception(f"Failed to load station: {ex}")
                raise ex
        else:
            try:
                structure = esi_fetch(
                    "Universe.get_universe_structures_structure_id",
                    args={"structure_id": location_id},
                    token=token,
                )

                eve_solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
                    id=structure["solar_system_id"]
                )

                location, created = self.update_or_create(
                    id=location_id,
                    defaults={
                        "name": structure["name"],
                        "eve_solar_system": eve_solar_system,
                        "category_id": Location.CATEGORY_STRUCTURE_ID,
                    },
                )
            except (HTTPUnauthorized, HTTPForbidden) as ex:
                logger.warning(f"No access to this structure: {ex}")
                if add_unknown:
                    location, created = self.get_or_create(
                        id=location_id,
                        defaults={
                            "name": f"Unknown structure {location_id}",
                            "category_id": Location.CATEGORY_STRUCTURE_ID,
                        },
                    )
                else:
                    raise ex
            except Exception as ex:
                logger.exception(f"Failed to load structure: {ex}")
                raise ex

        return location, created



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("buybacks2", "0002_programs"),
    ]

    operations = [
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "program_location",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="buybacks2.programlocation",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="auth.user",
                    ),
                ),
                (
                    "total",
                    models.PositiveBigIntegerField(
                        help_text="Total value of contract",
                    ),
                ),
                (
                    "items",
                    models.TextField(
                        help_text="JSON dump of item data",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Contract",
            fields=[
                (
                    "id",
                    models.PositiveBigIntegerField(
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "program",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="buybacks2.program",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
                (
                    "total",
                    models.PositiveBigIntegerField(
                        help_text="Total value of contract",
                    ),
                ),
                (
                    "date",
                    models.DateTimeField(),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render

from ..models import Contract


@login_required
@permission_required("buybacks2.basic_access")
def my_stats(request):
    contracts = Contract.objects.filter(
        character__user=request.user,
    )

    context = {
        "contracts": contracts,
        "mine": True,
    }

    return render(request, "buybacks2/stats.html", context)


@login_required
@permission_required("buybacks2.basic_access")
def program_stats(request, program_pk):
    contracts = Contract.objects.filter(
        program__pk=program_pk,
    )

    context = {
        "contracts": contracts,
        "mine": False,
    }

    return render(request, "buybacks2/stats.html", context)



import math
import re

from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from esi.decorators import token_required
from esi.models import Token

from ..forms import CalculatorForm, ProgramForm, ProgramItemForm, ProgramLocationForm
from ..helpers import fuzzworkmarket_lookup
from ..models import Corporation, Program, ProgramItem, ProgramLocation
from ..utils import MessagesPlus

ADD_PROGRAM_TOKEN_TAG = "buybacks_add_program_token"


@login_required
@permission_required("buybacks2.basic_access")
def program_calculate(request, program_pk):
    bb_program = Program.objects.filter(pk=program_pk).first()
    data = dict()
    value = {}
    typeids = {}
    program_location = None
    total = 0

    if bb_program is None:
        return redirect("buybacks2:index")

    if request.method != "POST":
        form = CalculatorForm(program=bb_program)
    else:
        form = CalculatorForm(request.POST, program=bb_program)

        if form.is_valid():
            office = form.cleaned_data["office"]
            items = form.cleaned_data["items"]

            program_location = ProgramLocation.objects.filter(
                program=bb_program,
                office__location__name=office,
            ).first()

            # Fix From https://gitlab.com/pksunkara/aa-buybacks/-/issues/34 - https://gitlab.com/Wered
            for line in items.splitlines():
                line = " ".join(line.split()).replace(",", "")
                filtered = re.match(r"(.*?)([0-9]+)", line).groups
                name = filtered("1")[0].rstrip()
                quantity = filtered("1")[1]
                if name in data:
                    data[name] += int(quantity)
                else:
                    data[name] = int(quantity)

            brokerages = {}

            program_items = ProgramItem.objects.filter(
                program=bb_program, item_type__name__in=data.keys()
            )

            for item in program_items:
                if not item.use_refined_value:
                    brokerages[item.item_type.name] = item.brokerage
                    typeids[item.item_type.name] = item.item_type.id

            prices = fuzzworkmarket_lookup(typeids.values())

            for name in brokerages:
                item_value = (
                    (100 - brokerages[name])
                    / 100
                    * data[name]
                    * prices[typeids[name]]["buy"]
                )

                value[name]["unit_price"] = item_value
                value[name]["typeid"] = typeids[name]
                value[name]["quantity"] = data[name]
                value[name]["name"] = name
                total += item_value

            total = math.ceil(total)

    context = {
        "program": bb_program,
        "corporation": bb_program.corporation.corporation,
        "form": form,
        "value": value,
        "total": total,
        "program_location": program_location,
    }

    return render(request, "buybacks2/program_calculate.html", context)


@login_required
@permission_required("buybacks2.basic_access")
def program(request, program_pk):
    bb_program = Program.objects.filter(pk=program_pk).first()

    if bb_program is None:
        return redirect("buybacks2:index")

    context = {
        "program": bb_program,
        "items": ProgramItem.objects.filter(program=bb_program),
        "locations": ProgramLocation.objects.filter(program=bb_program),
        "corporation": bb_program.corporation.corporation,
    }

    return render(request, "buybacks2/program.html", context)


@login_required
@permission_required("buybacks2.manage_programs")
def program_remove(request, program_pk):
    Program.objects.filter(pk=program_pk).delete()

    return redirect("buybacks2:index")


@login_required
@permission_required("buybacks2.manage_programs")
def program_remove_item(request, program_pk, item_type_pk):
    ProgramItem.objects.filter(
        program=program_pk,
        item_type=item_type_pk,
    ).delete()

    return redirect("buybacks2:program", program_pk=program_pk)


@login_required
@permission_required("buybacks2.manage_programs")
def program_remove_location(request, program_pk, office_pk):
    ProgramLocation.objects.filter(
        program=program_pk,
        office=office_pk,
    ).delete()

    return redirect("buybacks2:program", program_pk=program_pk)


@login_required
@permission_required("buybacks2.manage_programs")
def program_add_item(request, program_pk):
    bb_program = Program.objects.filter(pk=program_pk).first()

    if bb_program is None:
        return redirect("buybacks2:index")

    if request.method != "POST":
        form = ProgramItemForm()
    else:
        form = ProgramItemForm(
            request.POST,
            value=int(request.POST["item_type"]),
        )

        if form.is_valid():
            item_type = form.cleaned_data["item_type"]
            brokerage = form.cleaned_data["brokerage"]
            use_refined_value = False
            # use_refined_value = form.cleaned_data["use_refined_value"]

            try:
                _, created = ProgramItem.objects.update_or_create(
                    item_type=item_type,
                    program=bb_program,
                    defaults={
                        "brokerage": brokerage,
                        "use_refined_value": use_refined_value,
                    },
                )

                if created:
                    MessagesPlus.success(
                        request,
                        format_html(
                            "Added <strong>{}</strong> to <strong>{}</strong>",
                            item_type,
                            bb_program.name,
                        ),
                    )

                return redirect("buybacks2:program", program_pk=bb_program.id)

            except Exception:
                MessagesPlus.error(
                    request,
                    "Failed to add item to buyback program",
                )

    context = {
        "program": bb_program,
        "corporation": bb_program.corporation.corporation,
        "form": form,
    }

    return render(request, "buybacks2/program_add_item.html", context)


@login_required
@permission_required("buybacks2.manage_programs")
def program_add_location(request, program_pk):
    bb_program = Program.objects.filter(pk=program_pk).first()

    if bb_program is None:
        return redirect("buybacks2:index")

    if request.method != "POST":
        form = ProgramLocationForm(program=bb_program)
    else:
        form = ProgramLocationForm(request.POST, program=bb_program)

        if form.is_valid():
            office = form.cleaned_data["office"]

            try:
                _, created = ProgramLocation.objects.update_or_create(
                    office=office, program=bb_program
                )

                if created:
                    MessagesPlus.success(
                        request,
                        format_html(
                            "Added <strong>{}</strong> to <strong>{}</strong>",
                            office,
                            bb_program.name,
                        ),
                    )

                return redirect("buybacks2:program", program_pk=bb_program.id)

            except Exception:
                MessagesPlus.error(
                    request,
                    "Failed to add location to buyback program",
                )

    context = {
        "program": bb_program,
        "corporation": bb_program.corporation.corporation,
        "form": form,
    }

    return render(request, "buybacks2/program_add_location.html", context)


@login_required
@token_required(
    scopes=[
        "publicData",
    ]
)
@permission_required("buybacks2.manage_programs")
def program_add(request, token):
    request.session[ADD_PROGRAM_TOKEN_TAG] = token.pk
    return redirect("buybacks2:program_add_2")


@login_required
@permission_required("buybacks2.manage_programs")
def program_add_2(request):
    if ADD_PROGRAM_TOKEN_TAG not in request.session:
        raise RuntimeError("Missing token in session")
    else:
        token = Token.objects.get(pk=request.session[ADD_PROGRAM_TOKEN_TAG])

    success = True
    corporation = None
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    try:
        corporation = Corporation.objects.get(corporation=token_char.corporation)
    except (Corporation.DoesNotExist, EveCorporationInfo.DoesNotExist):
        MessagesPlus.error(
            request,
            format_html(
                gettext_lazy(
                    "You need to setup your corp first before managing it's buyback programs"
                )
            ),
        )
        success = False

    if success:
        if request.method != "POST":
            form = ProgramForm()
        else:
            form = ProgramForm(request.POST)

            if form.is_valid():
                name = form.cleaned_data["name"]

                try:
                    bb_program = Program.objects.create(
                        name=name, corporation=corporation
                    )
                    MessagesPlus.success(
                        request,
                        format_html(
                            "Created buyback program <strong>{}</strong>",
                            bb_program.name,
                        ),
                    )
                    return redirect("buybacks2:program", program_pk=bb_program.id)

                except Exception:
                    MessagesPlus.error(
                        request,
                        "Failed to create buyback program",
                    )

        context = {
            "corporation": corporation.corporation,
            "form": form,
            "to": reverse("buybacks2:program_add_2"),
            "action": "Create",
        }

        return render(request, "buybacks2/program_add.html", context)

    return redirect("buybacks2:index")


@login_required
@permission_required("buybacks2.manage_programs")
def program_edit(request, program_pk):
    bb_program = Program.objects.filter(pk=program_pk).first()

    if bb_program is None:
        return redirect("buybacks2:index")

    if request.method != "POST":
        form = ProgramForm(program=bb_program)
    else:
        form = ProgramForm(request.POST, program=bb_program)

        if form.is_valid():
            bb_program.name = form.cleaned_data["name"]

            try:
                bb_program.save()
                MessagesPlus.success(
                    request,
                    format_html(
                        "Edited buyback program <strong>{}</strong>",
                        bb_program.name,
                    ),
                )
                return redirect("buybacks2:program", program_pk=bb_program.id)

            except Exception:
                MessagesPlus.error(
                    request,
                    "Failed to edit buyback program",
                )

    context = {
        "corporation": bb_program.corporation.corporation,
        "form": form,
        "to": reverse(
            "buybacks2:program_edit",
            kwargs={
                "program_pk": program_pk,
            },
        ),
        "action": "Edit",
    }

    return render(request, "buybacks2/program_add.html", context)



from django import template

register = template.Library()


@register.filter
def format_number(value, p=1):
    """return a formated number with thousands seperators"""
    try:
        return f"{float(value):,.{int(p)}f}"
    except (ValueError, TypeError):
        return None



from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy


def validate_brokerage(value):
    if value > 100 or value < 0:
        raise ValidationError(
            gettext_lazy("%(value)s is not a valid brokerage"),
            params={"value": value},
        )



from django.contrib import admin

from .models import Corporation, Location, Office


@admin.register(Corporation)
class CorporationAdmin(admin.ModelAdmin):
    list_display = ("character", "corporation")

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Location)
class LocationAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "eve_solar_system")

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Office)
class OfficeAdmin(admin.ModelAdmin):
    list_display = ("id", "corporation", "location")

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



# Generated by Django 3.1.4 on 2020-12-26 17:11

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveuniverse", "0004_effect_longer_name"),
        ("authentication", "0017_remove_fleetup_permission"),
        ("eveonline", "0012_index_additions"),
    ]

    operations = [
        migrations.CreateModel(
            name="Buybacks",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app"),
                    ("setup_retriever", "Can setup information retriever"),
                    ("manage_programs", "Can manage buyback programs"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Corporation",
            fields=[
                (
                    "corporation",
                    models.OneToOneField(
                        primary_key=True,
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        help_text="Character used for retrieving info",
                        on_delete=models.deletion.PROTECT,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Location",
            fields=[
                (
                    "id",
                    models.PositiveBigIntegerField(
                        help_text="Eve Online location ID",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="In-game name of this station or structure",
                        max_length=100,
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.evesolarsystem",
                    ),
                ),
                (
                    "category_id",
                    models.IntegerField(
                        choices=[
                            (3, "station"),
                            (65, "structure"),
                            (0, "(unknown)"),
                        ],
                        default=0,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Office",
            fields=[
                (
                    "id",
                    models.PositiveBigIntegerField(
                        help_text="Item ID of the office",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="buybacks2.corporation",
                    ),
                ),
                (
                    "location",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="buybacks2.location",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
    ]



from json import loads

from django import template

register = template.Library()


@register.filter
def lookup(dictionary, key):
    return dictionary.get(key)


@register.filter
def not_empty(dictionary):
    return len(dictionary.keys()) != 0


@register.filter
def entries(dictionary):
    return dictionary.items()


@register.filter
def json(text):
    return loads(text)



from django.db import migrations, models

from ..validators import validate_brokerage


class Migration(migrations.Migration):

    dependencies = [
        ("buybacks2", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Program",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="buybacks2.corporation",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        max_length=100,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="ProgramItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "program",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="buybacks2.program",
                    ),
                ),
                (
                    "item_type",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "brokerage",
                    models.IntegerField(
                        help_text="Jita max buy - x%",
                        validators=[validate_brokerage],
                    ),
                ),
                (
                    "use_refined_value",
                    models.BooleanField(
                        help_text="If ore, calculate on top of refined value",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="ProgramLocation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "program",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="buybacks2.program",
                    ),
                ),
                (
                    "office",
                    models.ForeignKey(
                        on_delete=models.deletion.CASCADE,
                        related_name="+",
                        to="buybacks2.office",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.AlterUniqueTogether(
            name="ProgramItem",
            unique_together={("program", "item_type")},
        ),
        migrations.AlterUniqueTogether(
            name="ProgramLocation",
            unique_together={("program", "office")},
        ),
    ]



import json
from typing import Tuple

from django.contrib.auth.models import User
from django.db import models

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveSolarSystem, EveType

from .helpers import esi_fetch
from .managers import LocationManager
from .validators import validate_brokerage

logger = get_extension_logger(__name__)

OFFICE_TYPE_ID = 27


class Buybacks(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app"),
            ("setup_retriever", "Can setup information retriever"),
            ("manage_programs", "Can manage buyback programs"),
        )


class Corporation(models.Model):
    """A corporation that has buyback programs"""

    ERROR_NONE = 0
    ERROR_TOKEN_INVALID = 1
    ERROR_TOKEN_EXPIRED = 2
    ERROR_ESI_UNAVAILABLE = 5

    ERRORS_LIST = [
        (ERROR_NONE, "No error"),
        (ERROR_TOKEN_INVALID, "Invalid token"),
        (ERROR_TOKEN_EXPIRED, "Expired token"),
        (ERROR_ESI_UNAVAILABLE, "ESI API is currently unavailable"),
    ]

    corporation = models.OneToOneField(
        EveCorporationInfo,
        primary_key=True,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    character = models.ForeignKey(
        CharacterOwnership,
        help_text="Character used for retrieving info",
        on_delete=models.deletion.PROTECT,
        related_name="+",
    )

    class Meta:
        default_permissions = ()

    def sync_contracts_esi(self):
        token = self.token(
            [
                "esi-contracts.read_corporation_contracts.v1",
            ]
        )[0]

        contracts = esi_fetch(
            "Contracts.get_corporations_corporation_id_contracts",
            args={
                "corporation_id": self.corporation.corporation_id,
            },
            has_pages=True,
            token=token,
        )

        buybacks = [
            x
            for x in contracts
            if x["type"] == "item_exchange"
            and x["status"] == "finished"
            and int(x["assignee_id"]) == self.corporation.corporation_id
        ]

        for contract in buybacks:
            notification = Notification.objects.filter(
                program_location__office__location__id=contract["start_location_id"],
                total=contract["price"],
                user__character_ownerships__character__character_id=contract[
                    "issuer_id"
                ],
            ).first()

            if notification is not None:
                items = esi_fetch(
                    "Contracts.get_corporations_corporation_id_contracts_contract_id_items",
                    args={
                        "corporation_id": self.corporation.corporation_id,
                        "contract_id": contract["contract_id"],
                    },
                    token=token,
                )

                quantities = {}

                for item in items:
                    if item["is_included"]:
                        type_id = int(item["type_id"])
                        quantity = int(item["quantity"])

                        if type_id in quantities:
                            quantities[type_id] += quantity
                        else:
                            quantities[type_id] = quantity

                data = json.loads(notification.items)
                match = True

                for type_id in data:
                    type_id = int(type_id)

                    if (
                        type_id not in quantities
                        or quantities[type_id] != data[str(type_id)]
                    ):
                        match = False

                if match:
                    character = CharacterOwnership.objects.filter(
                        character__character_id=contract["issuer_id"]
                    ).first()

                    if character is not None:
                        Notification.objects.filter(
                            pk=notification.id,
                        ).delete()

                        Contract.objects.create(
                            id=contract["contract_id"],
                            program=notification.program_location.program,
                            character=character,
                            total=contract["price"],
                            date=contract["date_issued"],
                        )

    def update_offices_esi(self):
        token = self.token(
            [
                "esi-universe.read_structures.v1",
                "esi-assets.read_corporation_assets.v1",
            ]
        )[0]

        assets = esi_fetch(
            "Assets.get_corporations_corporation_id_assets",
            args={
                "corporation_id": self.corporation.corporation_id,
            },
            has_pages=True,
            token=token,
        )

        office_ids_to_remove = list(
            Office.objects.filter(corporation=self).values_list("id", flat=True)
        )

        for asset in assets:
            if asset["type_id"] == OFFICE_TYPE_ID:
                location = Location.objects.get_or_create_from_esi(
                    location_id=asset["location_id"],
                    token=token,
                )[0]

                office_id = asset["item_id"]
                office = Office.objects.filter(pk=office_id).first()

                if office is not None:
                    office_ids_to_remove.remove(office.id)
                else:
                    Office.objects.create(
                        id=office_id,
                        corporation=self,
                        location=location,
                    )

        Office.objects.filter(pk__in=office_ids_to_remove).delete()

    def token(self, scopes=None) -> Tuple[Token, int]:
        """returns a valid Token for the character"""
        token = None
        error = None

        try:
            # get token
            token = (
                Token.objects.filter(
                    user=self.character.user,
                    character_id=self.character.character.character_id,
                )
                .require_scopes(scopes)
                .require_valid()
                .first()
            )
        except TokenInvalidError:
            logger.error("Invalid token for fetching information")
            error = self.ERROR_TOKEN_INVALID
        except TokenExpiredError:
            logger.error("Token expired for fetching information")
            error = self.ERROR_TOKEN_EXPIRED
        else:
            if not token:
                logger.error("No token found with sufficient scopes")
                error = self.ERROR_TOKEN_INVALID

        return token, error

    def __str__(self):
        return self.corporation.corporation_name


class Location(models.Model):
    """An Eve Online buyback location: station or Upwell structure"""

    CATEGORY_STATION_ID = 3
    CATEGORY_STRUCTURE_ID = 65
    CATEGORY_UNKNOWN_ID = 0
    CATEGORY_CHOICES = [
        (CATEGORY_STATION_ID, "station"),
        (CATEGORY_STRUCTURE_ID, "structure"),
        (CATEGORY_UNKNOWN_ID, "(unknown)"),
    ]

    id = models.PositiveBigIntegerField(
        help_text="Eve Online location ID",
        primary_key=True,
    )
    name = models.CharField(
        max_length=100, help_text="In-game name of this station or structure"
    )
    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        blank=True,
        default=None,
        null=True,
        on_delete=models.deletion.SET_DEFAULT,
        related_name="+",
    )
    category_id = models.IntegerField(
        choices=CATEGORY_CHOICES,
        default=CATEGORY_UNKNOWN_ID,
    )

    objects = LocationManager()

    class Meta:
        default_permissions = ()

    def __str__(self):
        return self.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pk={self.pk}, name='{self.name}')"

    @property
    def category(self):
        return self.category_id

    @property
    def solar_system_name(self):
        return self.name.split(" ", 1)[0]

    @property
    def location_name(self):
        return self.name.rsplit("-", 1)[1].strip()


class Office(models.Model):
    """An Eve Online buyback office for a corp: station or Upwell structure"""

    id = models.PositiveBigIntegerField(
        primary_key=True,
        help_text="Item ID of the office",
    )
    corporation = models.ForeignKey(
        Corporation,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    location = models.ForeignKey(
        Location,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )

    class Meta:
        default_permissions = ()

    def __str__(self):
        return self.location.name


class Program(models.Model):
    """An Eve Online buyback program for a corp"""

    id = models.AutoField(
        auto_created=True,
        primary_key=True,
        verbose_name="ID",
    )
    corporation = models.ForeignKey(
        Corporation,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    name = models.CharField(
        max_length=100,
    )

    class Meta:
        default_permissions = ()


class ProgramItem(models.Model):
    """Items in the buyback program for a corp"""

    id = models.AutoField(
        auto_created=True,
        primary_key=True,
        verbose_name="ID",
    )
    program = models.ForeignKey(
        Program,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    item_type = models.ForeignKey(
        EveType,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    brokerage = models.IntegerField(
        help_text="Jita max buy - x%",
        validators=[validate_brokerage],
    )
    use_refined_value = models.BooleanField(
        help_text="If ore, calculate on top of refined value",
    )

    class Meta:
        default_permissions = ()
        unique_together = ["program", "item_type"]


class ProgramLocation(models.Model):
    """Locations for buyback program for a corp"""

    id = models.AutoField(
        auto_created=True,
        primary_key=True,
        verbose_name="ID",
    )
    program = models.ForeignKey(
        Program,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    office = models.ForeignKey(
        Office,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )

    def __str__(self):
        return self.office.location.name

    class Meta:
        default_permissions = ()
        unique_together = ["program", "office"]


class Notification(models.Model):
    """Notifiaction by user that he is sending contract"""

    id = models.AutoField(
        auto_created=True,
        primary_key=True,
        verbose_name="ID",
    )
    program_location = models.ForeignKey(
        ProgramLocation,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    user = models.ForeignKey(
        User,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    total = models.PositiveBigIntegerField(
        help_text="Total value of contract",
    )
    items = models.TextField(
        help_text="JSON dump of item data",
    )

    class Meta:
        default_permissions = ()


class Contract(models.Model):
    """Contract that is accepted in the buyback program"""

    id = models.PositiveBigIntegerField(
        primary_key=True,
    )
    program = models.ForeignKey(
        Program,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    character = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    total = models.PositiveBigIntegerField(
        help_text="Total value of contract",
    )
    date = models.DateTimeField()

    class Meta:
        default_permissions = ()



import logging

from django.conf import settings
from django.contrib import messages
from django.contrib.messages.constants import DEBUG, ERROR, INFO, SUCCESS, WARNING
from django.utils.html import format_html

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class LoggerAddTag(logging.LoggerAdapter):
    """
    add custom tag to a logger
    """

    def __init__(self, my_logger, prefix):
        super().__init__(my_logger, {})
        self.prefix = prefix

    def process(self, msg, kwargs):
        """
        process log items
        :param msg:
        :param kwargs:
        :return:
        """

        return f"[{self.prefix}] {msg}", kwargs


def make_logger_prefix(tag: str):
    """creates a function to add logger prefix, which returns tag with empty"""
    return lambda text="": f"{tag}{(': ' + text) if text else ''}"


class MessagesPlus:
    """Pendant to Django messages adding level icons and HTML support

    Careful: Use with safe strings only
    """

    _glyph_map = {
        DEBUG: "eye-open",
        INFO: "info-sign",
        SUCCESS: "ok-sign",
        WARNING: "exclamation-sign",
        ERROR: "alert",
    }

    @classmethod
    def _add_messages_icon(cls, level: int, message: str) -> str:
        """Adds an level based icon to standard Django messages"""
        if level not in cls._glyph_map:
            raise ValueError("glyph for level not defined")
        else:
            glyph = cls._glyph_map[level]

        return format_html(
            '<span class="glyphicon glyphicon-{}" '
            'aria-hidden="true"></span>&nbsp;&nbsp;{}',
            glyph,
            message,
        )

    @classmethod
    def debug(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.debug(
            request, cls._add_messages_icon(DEBUG, message), extra_tags, fail_silently
        )

    @classmethod
    def info(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.info(
            request, cls._add_messages_icon(INFO, message), extra_tags, fail_silently
        )

    @classmethod
    def success(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.success(
            request, cls._add_messages_icon(SUCCESS, message), extra_tags, fail_silently
        )

    @classmethod
    def warning(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.warning(
            request, cls._add_messages_icon(WARNING, message), extra_tags, fail_silently
        )

    @classmethod
    def error(
        cls,
        request: object,
        message: str,
        extra_tags: str = "",
        fail_silently: bool = False,
    ):
        messages.error(
            request, cls._add_messages_icon(ERROR, message), extra_tags, fail_silently
        )


def clean_setting(
    name: str,
    default_value: object,
    min_value: int = None,
    max_value: int = None,
    required_type: type = None,
):
    """cleans the input for a custom setting

    Will use `default_value` if settings does not exit or has the wrong type
    or is outside define boundaries (for int only)

    Need to define `required_type` if `default_value` is `None`

    Will assume `min_value` of 0 for int (can be overriden)

    Returns cleaned value for setting
    """
    if default_value is None and not required_type:
        raise ValueError("You must specify a required_type for None defaults")

    if not required_type:
        required_type = type(default_value)

    if min_value is None and required_type == int:
        min_value = 0

    if not hasattr(settings, name):
        cleaned_value = default_value
    else:
        if (
            isinstance(getattr(settings, name), required_type)
            and (min_value is None or getattr(settings, name) >= min_value)
            and (max_value is None or getattr(settings, name) <= max_value)
        ):
            cleaned_value = getattr(settings, name)
        else:
            logger.warning(
                "You setting for %s it not valid. Please correct it. "
                "Using default for now: %s",
                name,
                default_value,
            )
            cleaned_value = default_value
    return cleaned_value
